# -----------------------------------------------------------------------------
# xtablelist.tcl ---
# -----------------------------------------------------------------------------
# (c) 2016, Johann Oberdorfer - Engineering Support | CAD | Software
#     johann.oberdorfer [at] gmail.com
#     www.johann-oberdorfer.eu
# -----------------------------------------------------------------------------
# This source file is distributed under the BSD license.
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#   See the BSD License for more details.
# -----------------------------------------------------------------------------
# Purpose:
#   A TclOO class template to extend tablelist functionality.
#   Might be usefull as a starting point.
# -----------------------------------------------------------------------------
# TclOO naming conventions:
#   public methods  - starts with lower case declaration names, whereas
#   private methods - starts with uppercase naming, so we use CamelCase ...
# -----------------------------------------------------------------------------
# Credits:
#   tablelist widget: Copyright (c) 2000-2015  Csaba Nemethi
#   tooltip:          Copyright (c) 1996-2007 Jeffrey Hobbs
#   autoscroll:       Copyright (c) 2003 Kevin B Kenny
#   icons used from:  Open Icon Library
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# widget commands:
#  xtablelist::xtablelist
#  xtablelist::configure
#  xtablelist::cget
#
#  xtablelist::... + all commands from original tablelist
#                  for more sub-command refer as well to the
#                  public class interface
#
# -----------------------------------------------------------------------------
# widget options
#
#  -xhideoption  boolean "yes",1 / "no",0 (default: "no"),
#                hide/show popup-menu
#                to manipulate tablelist column display
#                default is "yes": show relev. menu items
#
#  -xshowbuttons boolean (default: "no"),
#                show sidebar action buttons to manipulate tablelist content

#  -xsortoption  boolean (default: "no"),
#                hide/show popup-menu to modify tablelist column sort order
#                
#  -xfileread    boolean (default: "no"),
#                allows reading in a previously stored file,
#
#  -xfilesave    boolean (default: "no")
#                 allows to saving tablelist data to a file
#
#  -xfilereadpostcmd
#                file read post command (default: ""),
#                if defined, allows to e.g. set widget state of depending
#                action buttons in the caller program
#
#  -xdeletepostcmd
#                table entry delete post command (default: "") 
#
#  -xrefreshpostcmd
#                executed after a delete / refresh action takes place
#
#  -xselectpostcmd
#                specify a command to be executed after selecting
#                an item in the tablelist (default: "")
#
#  -xmoveovereffect
#                boolean (default: "no"),
#                enable visual "move over" effect
#
# virtual events:
#
#   <<TLExtDataModified>>     - delete, modified or sort action took place
#   <<TLExtDataModifiedUndo>> - undo, nothing changed so far!
# -----------------------------------------------------------------------------
# Hint:
#   some of the options need to be specified when creating the widget
#   and are not fully configureable later on once the widget has been
#   created (open issue to be implemented later on...)
# -----------------------------------------------------------------------------


# for development: try to find autoscroll, etc ...
set this_file [file normalize [file dirname [info script]]]

# where to find required packages...
# set auto_path [linsert $auto_path 0 [file join $this_file ".." "contrib" ]]

package require Tk
package require TclOO
package require tablelist_tile
package require tooltip
package require autoscroll


# !!! to-do !!!
# package require Cvs2Biff
# package require Debug
# package require Biff


package provide xtablelist 0.4


# workaround for aqua
if { [tk windowingsystem] eq "aqua" } {
	interp alias {} ttk::scrollbar {} ::scrollbar
}


namespace eval xtablelist {

	# the widget's interface declaration
	namespace export \
		addValues \
		deleteAllItems

	set this_dir   [file dirname [info script]]
	set image_dir  [file join $this_dir "images"]
	set image_file [file join $this_dir "ImageLib.tcl"]
	
	variable cnt 0
	
	proc LoadImages {image_dir {patterns {*.gif}}} {
		foreach p $patterns {
			foreach file [glob -nocomplain -directory $image_dir $p] {
				set img [file tail [file rootname $file]]
				if { ![info exists images($img)] } {
					set images($img) [image create photo -file $file]
				}
			}}
		return [array get images]
	}
	
	proc xtablelist {path args} {
		variable cnt
	
		# this is a tk-like wrapper around my... class so that
		# object creation works like other tk widgets

		set obj [XTableListClass create tmp${cnt} $path {*}$args]
		incr cnt

		# rename oldName newName
		rename $obj ::$path
		return $path
	}
	
}

# -----------------------------------------------------------------------------
# Tk widget 'a like interface declaration 
# -----------------------------------------------------------------------------

oo::class create XTableListClass {
	
	constructor {path args} {

		my variable tblwidget
		my variable widgetOptions
		my variable widgetCompounds
	
		# ---------------------------------------------------------------
		# read images from library file or alternatively one by one
		# ---------------------------------------------------------------
		set image_file $::xtablelist::image_file
		set image_dir $::xtablelist::image_dir
		
		if { [file exists $image_file] } {
			source $image_file
			array set widgetCompounds [array get images]
		} else {
			array set widgetCompounds [::tablelistbrowser::LoadImages \
					[file join $image_dir] {"*.gif" "*.png"}]
		}
		# ---------------------------------------------------------------
		
		array set widgetCompounds {
			undo_flag 0
			sortby -1
			undoList {}
			DELIM "@"
			default_dir ""
		}

		# declaration of additional (configureable) widget options
		array set widgetOptions {
			-xtabheader  {}
			-xmoveovereffect 1
			-xhideoption  0
			-xsortoption  0
			-xshowbuttons 0
			-xfileread    0
			-xfilesave	  0
			-xfilereadpostcmd ""
			-xdeletepostcmd   ""
			-xrefreshpostcmd  ""
			-xselectpostcmd   ""
		}
		
		# incorporate arguments to local widget options,
		# needs to be done before "Build" as some opt's
		# are infulencing the widget creation only

		array set widgetOptions $args

		switch -- $::tcl_platform(platform) {
			"windows"        { set widgetCompounds(default_dir) $::env(TEMP) }
			"unix" - default { set widgetCompounds(default_dir) $::env(HOME) }
		}

		set widgetCompounds(tmp_file) \
				[file join $widgetCompounds(default_dir) "xtablelist.tmp"]
		
		# we use a frame for this specific widget class
		set f [ttk::frame $path -class xtablelist]

		# we must rename the widget command
		# since it clashes with the object being created
		set widget ${path}_
		my Build $f
		rename $path $widget
		
		my configure {*}$args
	}
	
	destructor {
		# add a destructor to clean up the widget
		set w [namespace tail [self]]
		catch {bind $w <Destroy> {}}
		catch {destroy $w}
	}
	
	method cget { {opt "" }  } {
		my variable tblwidget
		my variable widgetOptions
		
		if { [string length $opt] == 0 } {
			return [array get widgetOptions]
		}
		if { [info exists widgetOptions($opt) ] } {
			return $widgetOptions($opt)
		}
		return [$tblwidget cget $opt]
	}
	
	method configure { args } {
		my variable tblwidget
		my variable widgetOptions
		
		if {[llength $args] == 0}  {
			
			# return all tablelist options
			set opt_list [$tblwidget configure]
			
			# as well as all custom options
			foreach xopt [array get widgetOptions] {
				lappend opt_list $xopt
			}
			return $opt_list
			
		} elseif {[llength $args] == 1}  {
			
			# return configuration value for this option
			set opt $args
			if { [info exists widgetOptions($opt) ] } {
				return $widgetOptions($opt)
			}
			return [$tblwidget cget $opt]
		}
		
		# error checking
		if {[expr {[llength $args]%2}] == 1}  {
			return -code error "value for \"[lindex $args end]\" missing"
		}
		
		# process the new configuration options...
		array set opts $args
		
		foreach opt_name [array names opts] {
			set opt_value $opts($opt_name)
			
			# overwrite with new value
			if { [info exists widgetOptions($opt_name)] } {
				set widgetOptions($opt_name) $opt_value
			}
			
			# some options need action from the widgets side
			switch -- $opt_name {

				-labelcommand {
					# replace sortcolumn command by our own method...

					if { [my GetBoolean $widgetOptions(-xsortoption)] } {
						$tblwidget configure \
							-labelcommand "[namespace code {my SortByColumnCmd}]"
					} else {
						$tblwidget configure -labelcommand $opt_value
					}
				}
			
				-xtabheader {
					my InitializeTabHeader $opt_value
				}
				-xmoveovereffect  {
					if { [my GetBoolean $widgetOptions(-xmoveovereffect)] } {
						my EnableMoveover
					} else {
						$tblwidget configure -activestyle none
					}
				}

				-xhideoption {}
				-xsortoption {}
				-xshowbuttons {}
				-xfileread {}
				-xfilesave {}
				-xfilereadpostcmd {}
				-xdeletepostcmd {}
				-xrefreshpostcmd {}
				-xselectpostcmd {}
			
				default {

					# -------------------------------------------------------
					# if the configure option wasn't one of our special one's,
					# pass control over to the original tablelist widget
					# -------------------------------------------------------
					
					if {[catch {$tblwidget configure $opt_name $opt_value} result]} {
						return -code error $result
					}
				}
			}
		}
	}
	
	method insert {method args} {
		# overloaded tablelist method

		my variable tblwidget
		my variable widgetCompounds

		# maintain undo list + button states...
		lappend widgetCompounds(undoList) {*}$args

		# hand over to tablelist widget...
		if {[catch {$tblwidget insert $method {*}$args} result]} {
			return -code error $result
		}
		
		my UpdateStatusInfo [$tblwidget get 0 end]
	}


	method insertlist {method args} {
		# overloaded tablelist method

		my variable tblwidget

		# maintain undo list + button states...
		lappend widgetCompounds(undoList) {*}$args
			
		if {[catch {$tblwidget insertlist $method {*}$args} result]} {
			return -code error $result
		}
		
		my UpdateStatusInfo [$tblwidget get 0 end]
	}


	method unknown {method args} {
	
		# if the command wasn't one of our special one's,
		# pass control over to the original tablelist widget

		my variable tblwidget
		
		if {[catch {$tblwidget $method {*}$args} result]} {
			return -code error $result
		}
		return $result
	}
}


# -----------------------------------------------------------------------------
# Private interface methods
# -----------------------------------------------------------------------------

oo::define XTableListClass {

	method SortByColumnCmd {args} {
		my variable tblwidget
		my variable widgetCompounds
	
		# set tbl [lindex $args 0]
		set idx [lindex $args 1]
	
		set widgetCompounds(sortby) $idx
		::tablelist::sortByColumn $tblwidget $idx
	}

	method InitBindings {} {
		my variable tblwidget
		
		bind [$tblwidget bodypath] <ButtonPress-3> {
			set t [winfo parent %W]
			set x [expr {%x + [winfo x %W]}]
			set y [expr {%y + [winfo y %W]}]
			
			set oo_my [winfo parent $t]
			
			# move up hierachy until our own class is reached:
			# notes:
			#   - maybe there is a more convinient way to adress a class member function ?
			#   - pls note that the member function needs to start with lowercase char
			#     otherwise it's hidden in global namespace
			
			while { [winfo exists $oo_my] } {
				# puts "--> $oo_my : [winfo class $oo_my]"
				
				if { [winfo class $oo_my] == "xtablelist" } { break }
				set oo_my [winfo parent $oo_my]
			}
			
			$oo_my showContextMenu \
					[$t nearestcolumn $x] [$t nearest $y] %X %Y
		}
	}
	
	method Build {frm} {

		# test...
		# extract the "widget name" from the object name:
		# set w [namespace tail [self]]

		# build the (mega-) widget
		# frm - parent frame widget path

		my variable tblwidget
		my variable widgetOptions
		my variable widgetCompounds
		
		# status line
		# -----------
		set f [ttk::frame $frm.flabel]
		pack $f -side bottom -fill both
		
		ttk::label $f.lbl -text ""
		
		pack $f.lbl -side bottom -fill x -anchor w
		
		set widgetCompounds(label) $f.lbl
		
		# listbox and two scrollbars embedded into a frame
		
		set f [ttk::frame $frm.f]
		pack $f -side left -fill both -expand true
		
		# a.) right
		set fright [ttk::frame $f.fright]
		pack $fright -side right -fill y
		
		ttk::scrollbar $fright.scrolly -orient vertical
		pack $fright.scrolly -side top -fill y -expand true
		
		ttk::frame $fright.spacery -height 15
		pack $fright.spacery -side bottom
		
		# b.) left
		set fleft [ttk::frame $f.fleft]
		pack $fleft -side left -fill both -expand true
		
		# left/top frame
		set fltop [ttk::frame $fleft.fltop]
		
		# left/bottom x scrollbar
		ttk::scrollbar $fleft.scrollx -orient horizontal
		
		# pack scrollbar 1st, so that it won't disappear while panning
		pack $fleft.scrollx -side bottom -fill x
		pack $fltop -side top -fill both -expand true
		
		# another frame for the tablelist object
		set fxx [ttk::frame $fltop.fleft]
		pack $fxx -side top -fill both -expand true
		
		::tablelist::tablelist $fxx.tlist \
				-selectmode extended \
				-takefocus 0 \
				-activestyle none \
				-xscrollcommand "$fleft.scrollx set" \
				-yscrollcommand "$fright.scrolly set"
		
		pack $fxx.tlist -side bottom -fill both -expand true
		
		set tblwidget $fxx.tlist

		bind $tblwidget <<TablelistSelect>> \
				"+ [namespace code {my SelectionCallback}]"
		
		bind $tblwidget <<TablelistCellUpdated>> \
				"+ event generate $tblwidget <<TLExtDataModified>>"
		
		
		$fleft.scrollx configure -command  "$fxx.tlist xview"
		$fright.scrolly configure -command "$fxx.tlist yview"
		
		# enable auto scrolling...
		autoscroll::autoscroll $fleft.scrollx
		autoscroll::autoscroll $fright.scrolly
		
		# icon-box for list manipulation ...
		# ----------------------------------
		set f1 [ttk::frame $frm.f1]
		pack $f1 -side right -fill both
		
		if { [my GetBoolean $widgetOptions(-xshowbuttons)] } {
			
			ttk::button $f1.b1 -image $widgetCompounds(up) \
					-style Toolbutton \
					-command "[namespace code {my ToolbuttonCallback}] 1"
			
			::tooltip::tooltip $f1.b1 "Move-Up"
			
			pack $f1.b1 -pady 1
			set widgetCompounds(b1) $f1.b1
			
			ttk::button $f1.b1i -image $widgetCompounds(reverselist) \
					-style Toolbutton \
					-command "[namespace code {my ToolbuttonCallback}] 10"
			
			::tooltip::tooltip $f1.b1i "Reverse the list order"
			
			pack $f1.b1i -pady 1
			set widgetCompounds(b1i) $f1.b1i
			
			ttk::button $f1.b2 -image $widgetCompounds(down) \
					-style Toolbutton \
					-command "[namespace code {my ToolbuttonCallback}] 2"
			
			::tooltip::tooltip $f1.b2 "Move-Down"
			
			pack $f1.b2 -pady 1
			set widgetCompounds(b2) $f1.b2
			
			
			ttk::button $f1.b3 -image $widgetCompounds(delete) \
					-style Toolbutton \
					-command "[namespace code {my ToolbuttonCallback}] 3"
			
			::tooltip::tooltip $f1.b3 "Remove item(s) from the list"
			
			pack $f1.b3 -pady 1
			set widgetCompounds(b3) $f1.b3
			
			ttk::button $f1.b3i -image $widgetCompounds(delete_reverse) \
					-style Toolbutton \
					-command "[namespace code {my ToolbuttonCallback}] 5"
			
			::tooltip::tooltip $f1.b3i "Remove All item(s)- except selected"
			
			pack $f1.b3i -pady 1
			set widgetCompounds(b3i) $f1.b3i
			
			
			ttk::button $f1.b4 -image $widgetCompounds(reload) \
					-style Toolbutton \
					-command "[namespace code {my ToolbuttonCallback}] 4"
			
			::tooltip::tooltip $f1.b4 "Undo previous operation"
			
			pack $f1.b4 -pady 1
			set widgetCompounds(b4) $f1.b4
			
			
		} elseif {$widgetOptions(-xshowbuttons) == "arrows"} {
			
			ttk::button $f1.b1 -image $widgetCompounds(up) \
					-style Toolbutton \
					-command "[namespace code {my ToolbuttonCallback}] 1"
			
			::tooltip::tooltip $f1.b1 "Move-Up"
			
			pack $f1.b1 -pady 1
			set widgetCompounds(b1) $f1.b1
			
			ttk::button $f1.b2 -image $widgetCompounds(down) \
					-style Toolbutton \
					-command "[namespace code {my ToolbuttonCallback}] 2"
			
			::tooltip::tooltip $f1.b2 "Move-Down"
			
			pack  $f1.b2 -pady 1
			set widgetCompounds(b2)  $f1.b2
		}
		
		
		if { [my GetBoolean $widgetOptions(-xfileread)] } {
			
			ttk::button $f1.b8 -image $widgetCompounds(folder) \
					-style Toolbutton \
					-command "[namespace code {my OpenFromFile}]"
			
			::tooltip::tooltip $f1.b8 "Open Listbox content from file"
			
			pack $f1.b8 -pady 1 -side bottom
			set widgetCompounds(b8)  $f1.b8
			
		}
		
		if { [my GetBoolean $widgetOptions(-xfilesave)] } {
			
			ttk::button $f1.b5 -image $widgetCompounds(filesaveas) \
					-style Toolbutton \
					-command "[namespace code {my Save2File}]"
			
			::tooltip::tooltip $f1.b5 "Save Listbox to file"
			
			pack $f1.b5 -pady 1 -side bottom
			set widgetCompounds(b7) $f1.b5
			
		}
		
		# disabled by default:
		if { [my GetBoolean $widgetOptions(-xshowbuttons)] } {
			
			$widgetCompounds(b1) configure -state disabled
			$widgetCompounds(b2) configure -state disabled
			$widgetCompounds(b3) configure -state disabled
			$widgetCompounds(b3i) configure -state disabled
			$widgetCompounds(b4) configure -state disabled
			
		} elseif {$widgetOptions(-xshowbuttons) == "arrows"} {
			$widgetCompounds(b1) configure -state disabled
			$widgetCompounds(b2) configure -state disabled
		}
		
		#$tlist($w,b6) configure -state disabled
		if { [my GetBoolean $widgetOptions(-xfilesave)] } {
			$widgetCompounds(b7) configure -state disabled
		}
		
		# note: There is also a popup-menu, which corresponds with
		#       the button box. So, if you do any changes here,
		#       do not forget to maintain there as well!
		set widgetCompounds(mstate) disabled
		
		my InitBindings
	}
	
}
