# -------------------------------------------------------------------------
# xtablelist_filesel.tcl ---
# -------------------------------------------------------------------------
#
# Purpose:
#   Utility package supporting a more convinient file selection dialog.
#   The following additional features where added:
#     - the possibility to generate a file-name,
#     - use the dialog's default file-name,
#       which might be sufficient for most of all cases,
#     - use standard file selection dialog
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
# file action utility dialog ...
# -------------------------------------------------------------------------

namespace eval xtablelist {
	
	variable widgetDefaults
	variable widgetImages
	
	switch -- $::tcl_platform(platform) {
		"windows"        { set homeDir $::env(TEMP) }
		"unix" - default { set homeDir $::env(HOME) }
	}
	
	array set widgetDefaults {
		title "Selection file:"
		DEFNAME "NONAME"
		def_ext ".txt"
		oktext " OK "
		ENTRY_WITH 45
		is_ok 0
	}

	set widgetDefaults(init_dir)    $homeDir
	set widgetDefaults(defaultfile) [file join $homeDir "NONAME.txt"]
	set widgetDefaults(filetypes)   {{"Text-Files" "*.txt*"} {all "*.*"}}
	
	# initializing required images...
	set imgdir [file join [file dirname [info script]] "images"]
	
	proc LoadImages1 {imgdir {patterns {*.gif *.png}}} {
		foreach pattern $patterns {
			foreach file [glob -nocomplain -directory $imgdir $pattern] {
				set img [file tail [file rootname $file]]
				if { ![info exists images($img)]} {
					set images($img) [image create photo -file $file]
				}
			}
		}
		return [array get images]
	}
	
	array set widgetImages [LoadImages1 $imgdir]
}


proc xtablelist::PositionWindow {} {
	variable widgetDefaults
	wm geometry $widgetDefaults(this) "+250+350"
}

proc xtablelist::_ok_callback {} {
	variable widgetDefaults
	
	if [file exists [set fname [$widgetDefaults(input_entry) get]]] {

		set answer [tk_messageBox \
						-parent $widgetDefaults(this) \
						-title "Question: " \
						-message "$fname\nalready exists\noverwrite ?" \
						-type yesno -icon question]

		if {$answer eq "no"} {
			set widgetDefaults(is_ok) 2
			return
		}
	}
	
	set widgetDefaults(is_ok) 1
}

proc xtablelist::EntryBindingsCmd {} {
	variable widgetDefaults
	variable widgetImages
	
	if {[$widgetDefaults(input_entry) get] == ""} {
		$widgetDefaults(input_label) configure -image $widgetImages(isempty)
		$widgetDefaults(ok_button) configure -state disabled
	} else {
		$widgetDefaults(input_label) configure -image $widgetImages(ok)
		$widgetDefaults(ok_button) configure -state normal
	}
}


proc xtablelist::SelectFileCmd {} {
	variable widgetDefaults
	variable widgetImages
	
	set log_file [tk_getSaveFile \
			-initialdir $widgetDefaults(init_dir) \
			-defaultextension $widgetDefaults(def_ext) \
			-filetypes  $widgetDefaults(filetypes) \
			-initialfile "" -title "Select file dialog..." \
			-parent $widgetDefaults(this)]
	
	if {$log_file != ""} {
		set widgetDefaults(init_dir) [file dirname $log_file]
		
		set widgetDefaults(entry_val) $log_file
		$widgetDefaults(input_label) configure -image $widgetImages(ok)
		$widgetDefaults(ok_button) configure -state normal
	}
}

proc xtablelist::GenerateFilename {} {
	variable widgetDefaults
	variable widgetImages
	
	# note: it is not possible under windows to use ":" as date separator!
	#       clock format [clock seconds] -format "%H:%M:%S"
	
	set fname [clock format [clock seconds] -format "%Y-%m-%d"]
	append fname "_"
	append fname $widgetDefaults(DEFNAME)
	append fname $widgetDefaults(def_ext)
	
	set widgetDefaults(entry_val) [file join $widgetDefaults(init_dir) $fname]
	$widgetDefaults(input_label) configure -image $widgetImages(ok)
}

proc xtablelist::CancelCmd {} {
	variable widgetDefaults
	set widgetDefaults(is_ok) 2
}


# -------------------------------------------------------------------------
# gui declaration
# -------------------------------------------------------------------------

proc xtablelist::getfilename {args} {
	variable widgetDefaults
	variable widgetImages
	
	set wparent ""
	set ind 0
	
	while { $ind < [llength $args] } {
		switch -exact -- [lindex $args $ind] {
			"-parent" {
				incr ind
				set wparent [lindex $args $ind]
				incr ind
			}
			"-title" {
				incr ind
				set widgetDefaults(title) [lindex $args $ind]
				incr ind
			}
			"-defaultfile" {
				incr ind
				set widgetDefaults(defaultfile) [lindex $args $ind]
				incr ind
			}
			"-defaultext" {
				incr ind
				set widgetDefaults(def_ext) [lindex $args $ind]
				incr ind
			}
			"-filetypes" {
				incr ind
				set widgetDefaults(filetypes) [lindex $args $ind]
				incr ind
			}
			"-oktext" {
				incr ind
				set widgetDefaults(oktext) [lindex $args $ind]
				incr ind
			}
			default {
				puts "unknown option [lindex $args $ind]"
				return ""
			}
		}
	}
	
	set w $wparent.getfilename
	set widgetDefaults(this) $w
	
	catch {destroy $w}
	toplevel $w
	wm title $w $widgetDefaults(title)
	wm iconname $w $widgetDefaults(title)
	PositionWindow
	wm transient $w $wparent
	bind $w <KeyPress-Escape> "[namespace current]::CancelCmd"
	
	set f [ttk::frame $w.info -relief flat]
	pack $f -side top -fill x -padx 1 -pady 1
	
	ttk::label $f.img -image $widgetImages(job_engineering_logo)
	
	set msg "
	Select a filename or press
	\"generate-filename\" to create one."
	
	ttk::label $f.msg -text $msg
	pack $f.img $f.msg -side left
	
	# button row
	# ----------
	set f [ttk::frame $w.buttons -relief groove]
	pack $f -side bottom -fill x -pady 2
	
	# image + button
	# --------------
	set fb [ttk::frame $f.img_button]
	# pack follows later on: pack $fb -side left -fill x -expand true
	
	ttk::button $fb.ok \
		-text $widgetDefaults(oktext) \
		-image $widgetImages(filesaveas) \
		-compound left \
		-command "[namespace current]::_ok_callback" \
		-state disabled

	pack $fb.ok -side left -padx 4 -pady 4
	set widgetDefaults(ok_button) $fb.ok

	ttk::button $f.cancel \
		-text "Cancel" \
		-command "[namespace current]::CancelCmd"

	pack $fb $f.cancel -side left -padx 4 -pady 4 -expand true
	
	
	set f [ttk::frame $w.rbuttons]
	pack $f -side bottom -fill x -padx 2 -pady 2
	
	ttk::button $f.getlatest \
		-text "Generate filename ..." \
		-command "[namespace current]::GenerateFilename" \
		-style Toolbutton

	pack $f.getlatest -padx 4 -pady 4
	
	set f_file1 [ttk::frame $f.part_num1]
	pack $f_file1 -fill x -expand true


	ttk::label $f_file1.img -image $widgetImages(isempty)
	pack $f_file1.img -side left
	set widgetDefaults(input_label) $f_file1.img

	pack [ttk::label $f_file1.lb -text "File-name:"] -side left
	
	ttk::entry $f_file1.en \
			-width $widgetDefaults(ENTRY_WITH) \
			-textvariable "[namespace current]::widgetDefaults(entry_val)"
	
	pack $f_file1.en -side left -fill x -padx 4 -expand true
	set widgetDefaults(input_entry) $f_file1.en

	ttk::button $f_file1.b \
		-text "..." \
		-command "[namespace current]::SelectFileCmd"
			
	pack $f_file1.b -side left
	
	bind $widgetDefaults(input_entry) <Leave>  "[namespace current]::EntryBindingsCmd"
	bind $widgetDefaults(input_entry) <Return> "[namespace current]::EntryBindingsCmd"
	
	set widgetDefaults(entry_val) $widgetDefaults(defaultfile)
	
	focus $widgetDefaults(input_entry)
	
	if {[$widgetDefaults(input_entry) get] != ""} {

		$widgetDefaults(input_label) configure -image $widgetImages(ok)
		$widgetDefaults(input_entry) selection range 0 end
		
		$widgetDefaults(ok_button) configure -state normal
	}
	
	# wait user
	grab $widgetDefaults(this)
	tkwait variable "[namespace current]::widgetDefaults(is_ok)"
	grab release $widgetDefaults(this)
	
	if { $widgetDefaults(is_ok) == 1 } {
		set retval [$widgetDefaults(input_entry) get]
	} else {
		set retval ""
	}
	
	destroy $w
	return $retval
}


# -------------------------------------------------------------------------
# test-run ...
# -------------------------------------------------------------------------
if 0 {
	package require Tk
	wm withdraw .
	catch {console show}

	set rval [xtablelist::getfilename \
			-parent      "" \
			-title       "Select file where to store Listbox information:" \
			-defaultfile [file join $env(HOME) "Test.txt"] \
			-filetypes   {{"Text-Files" "*.txt"} {all "*.*"}} \
			-defaultext  ".txt" \
			-oktext      " Save to File... "]

	puts "File selection result: \"$rval\""
}
