# -----------------------------------------------------------------------------
# xtablelist_private.tcl ---
# -----------------------------------------------------------------------------
# (c) 2016, Johann Oberdorfer - Engineering Support | CAD | Software
#     johann.oberdorfer [at] gmail.com
#     www.johann-oberdorfer.eu
# -----------------------------------------------------------------------------
# This source file is distributed under the BSD license.
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#   See the BSD License for more details.
# -----------------------------------------------------------------------------
# Purpose:
#  A TclOO class template to extend tablelist functionality.
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
# Private interface methods
# -----------------------------------------------------------------------------

oo::define XTableListClass {
	
	method GetBoolean {value {errorValue 1}} {
		
		# returns the value of a (presumably) boolean string (ie: it should
		# do the right thing if the string is "yes", "no", "true", 1, etc
		#
		# value - value to be converted
		# errorValue - a default value to be returned in case of an error
		#
		# Returns:
		#   a 1 or zero, or the value of errorValue if the string isn't
		#   a proper boolean value
		
		if {[catch {expr {([string trim $value])?1:0}} res]} {
			return $errorValue
		} else {
			return $res
		}
	}
	
	method GetTblColumnNames {} {
		# retrieve all column names, regardless if they are hidden or not
		my variable tblwidget

		set col_names {}
		foreach {with name pos} [$tblwidget cget -columns] {
			lappend col_names $name
		}
		return $col_names
	}

	method GetOnlyVisibleTblColumnNames {} {
		# retrieve only column names which are not hidden

		my variable tblwidget

		set cnt 0
		set col_names {}
		
		foreach {with name pos} [$tblwidget cget -columns] {
		
			if { ! [$tblwidget columncget $cnt -hide] } {
				lappend col_names $name
			}
			incr cnt
		}
		return $col_names
	}
	
	method InitializeTabHeader {kword_list} {
		my variable tblwidget
		
		set cols ""
		set cnt 0
		foreach i $kword_list {
			set descr  [lindex $i 0]
			set visual [lindex $i 1]
			set orient [lindex $i 2]
			
			if {[string length $orient] == 0} {
				set orient "left"
			}
			
			regsub -all " " $descr "_" descr
			if {$descr != "..." &&
				[string range $descr end end] != ":"} {
				set descr "${descr}:"
			}
			
			# could be either a string or an integer:
			if {$visual == "hidden"} {
				set width 20
			} else {
				set width $visual
			}
			
			append cols "$width $descr $orient "
			incr cnt
		}
		
		$tblwidget configure -columns $cols
		
		# hide specific columns as indicated with "hidden" in declaration array
		set cnt 0
		foreach i $kword_list {
			if {[set width [lindex $i 1]] == "hidden"} {
				$tblwidget columnconfigure $cnt -hide yes
			}
			incr cnt
		}
		
		# expand last *visible* column
		# ----------------------------
		set cnt 0
		set lastcol 0
		# tablelist -columns option are always 3 attributes each...
		foreach {w col pos} [$tblwidget cget -columns] {
			if {[$tblwidget columncget $cnt -hide] == 0} {set lastcol  $cnt}
			incr cnt
		}
		if {$lastcol > 0} {
			$tblwidget configure -stretch $lastcol
		}
	}

	method SelectionCallback {} {
		my variable tblwidget
		my variable widgetCompounds
		my variable widgetOptions
		
		if { ![llength [$tblwidget get 0 end]] } {
			return
		}
		
		if { ![my GetBoolean $widgetOptions(-xshowbuttons)] } {
			return
		}

		# manage visualisation mode of toolboxes
		set csel [$tblwidget curselection]
		
		if {$widgetOptions(-xshowbuttons) == "arrows"} {
			
			if {$csel == {}} {
				$widgetCompounds(b1) configure -state disabled
				$widgetCompounds(b2) configure -state disabled
			} else {
				$widgetCompounds(b1) configure -state normal
				$widgetCompounds(b2) configure -state normal
			}
			
		} else {
			
			if {$csel == {}} {
				$widgetCompounds(b1) configure -state disabled
				$widgetCompounds(b2) configure -state disabled
				$widgetCompounds(b3) configure -state disabled
				$widgetCompounds(b3i) configure -state disabled
			} else {
				$widgetCompounds(b1) configure -state normal
				$widgetCompounds(b2) configure -state normal
				$widgetCompounds(b3) configure -state normal
				$widgetCompounds(b3i) configure -state normal
			}
		}
		
		
		# does it make sense to enable undo button:
		if {$widgetCompounds(undo_flag) == 1} {
			
			$widgetCompounds(b4) configure -state normal
			event generate $tblwidget <<TLExtDataModified>>
			
		} else {
			
			$widgetCompounds(b4) configure -state disabled
			event generate $tblwidget <<TLExtDataModifiedUndo>>
		}
		
		
		# move-up
		if {[lsearch $csel 0] != -1} {
			$widgetCompounds(b1) configure -state disabled
		}
		
		# move-down
		set data [$tblwidget get 0 end]
		
		set len [expr [llength $data] -1]
		if {[lsearch $csel $len] != -1} {
			$widgetCompounds(b2) configure -state disabled
		}

		# evaluate post command in parent namespace...
		if {$widgetOptions(-xselectpostcmd) != ""} {
			catch {uplevel $widgetOptions(-xselectpostcmd)}
		}
	}

	
	method UpdateStatusInfo {lst} {

		# method to maintain additional status information
		
		my variable tblwidget
		my variable widgetCompounds
		my variable widgetOptions
		
		set msg ""
		
		switch -- [set cnt [llength $lst]] {
			0 {
				$widgetCompounds(label) configure -text "Tablelist is empty."

				if { [my GetBoolean $widgetOptions(-xfilesave)] } {
					$widgetCompounds(b7) configure -state disabled
				}
				
				# required to manage state, when the widget is popping up!
				set widgetCompounds(mstate) disabled
				
				# maintain pop-up menu
				# note: catch is required if the pop-up is not "pinned" on
				#       the screen (which is most likely the case!)

				if { [my GetBoolean $widgetOptions(-xhideoption)] } {
					set idx_list {8 9 10}
				} else {
					set idx_list {5 6 7}
				}

				if { [my GetBoolean $widgetOptions(-xshowbuttons)] } {
					$widgetCompounds(b1i) configure -state disabled
				}
				
				return
			}
			1 {
			
				if { [my GetBoolean $widgetOptions(-xshowbuttons)] } {
					$widgetCompounds(b1i) configure -state disabled
				}

				$widgetCompounds(label) configure -text "1 item in tablelist. $msg"
			}
			default {

				if { [my GetBoolean $widgetOptions(-xshowbuttons)] } {
					$widgetCompounds(b1i) configure -state normal
				}
				$widgetCompounds(label) configure -text "$cnt items in tablelist. $msg"
			}
		}
		
		if { [my GetBoolean $widgetOptions(-xfilesave)] } {
			$widgetCompounds(b7) configure -state normal
		}
		
		set widgetCompounds(mstate) normal

		if { [my GetBoolean $widgetOptions(-xhideoption)] } {

			set idx_list {8 9 10}
		} else {
			set idx_list {5 6 7}
		}

	}
	
	method ToolbuttonCallback {nr} {
		my variable tblwidget
		my variable widgetCompounds
		my variable widgetOptions
		
		# refresh
		
		if {$nr == 4} {
			$tblwidget delete 0 end
			eval $tblwidget insert end $widgetCompounds(undoList)
			
			set widgetCompounds(sortby) -1
			set widgetCompounds(undo_flag) 0
			
			my UpdateStatusInfo $widgetCompounds(undoList)
			
			# evaluate post command in parent namespace...
			if {$widgetOptions(-xrefreshpostcmd) != ""} {
				catch {uplevel $widgetOptions(-xrefreshpostcmd)}
			}
			
			my SelectionCallback
			return
		}
		
		# reverse the list
		if {$nr == 10} {
			
			set data [$tblwidget get 0 end]
			set data_len [llength $data]
			set csel [$tblwidget curselection]
			
			set reversed_data {}
			set reversed_csel {}
			
			for {set cnt [expr {$data_len -1}]} {$cnt >=  0} {incr cnt -1} {
				lappend reversed_data [lindex $data $cnt]
			}
			
			foreach idx $csel {
				lappend reversed_csel [expr {$data_len - $idx }]
			}
			
			$tblwidget delete 0 end
			eval $tblwidget insert end $reversed_data
			
			# highlight again ...
			foreach idx $reversed_csel {
				set idx1 [expr {$idx -1}]
				$tblwidget selection set $idx1 $idx1
			}
			
			# re-position listbox acc. to current selection
			set csel [$tblwidget curselection]
			$tblwidget see [lindex $csel 0]
			
			my selectArrowCallback
			
			set widgetCompounds(undo_flag) 1
		}
		
		set csel [$tblwidget curselection]
		
		if {$csel == {}} {
			
			# call get's important, if more than one widgets exists
			# at the same time, and the focus get's lost/shiftet:
			# action buttons are still enabled, but
			# the following call 'll disable them accordingly
			my SelectionCallback
			
			return
		}
		
		switch -- $nr {
			1 {
				# move-up
				if {[lsearch $csel 0] != -1} {
					return
				}
				
				set data [$tblwidget get 0 end]
				
				for {set cnt 0} {$cnt < [llength $data]} {incr cnt} {
					
					set cnt0 [expr $cnt -1]
					set tmp0 [lindex [lrange $data $cnt0 $cnt0] 0]
					set tmp  [lindex [lrange $data $cnt  $cnt] 0]
					
					if {[lsearch $csel $cnt] != -1} {
						set data [lreplace $data $cnt0 $cnt0 $tmp]
						set data [lreplace $data $cnt  $cnt  $tmp0]
					}
				}
				
				$tblwidget delete 0 end
				eval $tblwidget insert end $data
				
				# highlight again ...
				foreach idx $csel {
					set idx0 [expr $idx -1]
					$tblwidget selection set $idx0 $idx0
				}
				
				# 06.06.05: and (!) position listbox as it was before!
				$tblwidget see [lindex $csel 0]
				
				my selectArrowCallback
				
				set widgetCompounds(undo_flag) 1
			}
			2 {
				# move-down
				set data [$tblwidget get 0 end]
				set len [expr [llength $data] -1]
				
				if {[lsearch $csel $len] != -1} {
					return
				}
				
				set cnt 0
				for {set cnt [llength $data]} {$cnt >=  0} {incr cnt -1} {
					
					set cnt1 [expr $cnt +1]
					set tmp1 [lindex [lrange $data $cnt1 $cnt1] 0]
					set tmp  [lindex [lrange $data $cnt $cnt] 0]
					
					if {[lsearch $csel $cnt] != -1} {
						set data [lreplace $data $cnt  $cnt  $tmp1]
						set data [lreplace $data $cnt1 $cnt1 $tmp]
					}
				}
				
				$tblwidget delete 0 end
				eval $tblwidget insert end $data
				
				# highlight again ...
				foreach idx $csel {
					set idx1 [expr $idx +1]
					$tblwidget selection set $idx1 $idx1
				}
				
				# set visual region as it was before...
				$tblwidget see [lindex $csel 0]
				
				my selectArrowCallback
				
				set widgetCompounds(undo_flag) 1
			}
			3 {
				# does not work with empty lines:
				# foreach i [lsort -decreasing $csel] { $w delete $i $i }
				set data [$tblwidget get 0 end]
				
				set cnt 0
				for {set cnt [llength $data]} {$cnt >=  0} {incr cnt -1} {
					
					# check if index is within current selection
					# if true - remove list item !
					if {[lsearch $csel $cnt] != -1} {
						set data [lreplace $data $cnt $cnt]
					}
				}
				
				$tblwidget delete 0 end
				eval $tblwidget insert end $data
				
				my UpdateStatusInfo $data
				
				# evaluate post command in parent namespace...
				if {$widgetOptions(-xdeletepostcmd) != ""} {
					catch {uplevel $widgetOptions(-xdeletepostcmd)}
				}
				
				set widgetCompounds(undo_flag) 1
			}
			5 {
				# inverse delete = delete all, except selected
				set data [$tblwidget get 0 end]
				set new_data {}
				
				for {set cnt 0} {$cnt < [llength $data]} {incr cnt 1} {
					
					# check if the current index is *not* part of the selection ...
					if {[lsearch $csel $cnt] != -1} {
						lappend new_data [lindex $data $cnt]
						# puts "--> [lindex $data $cnt]"
					}
				}
				
				$tblwidget delete 0 end
				eval $tblwidget insert end $new_data
				
				my UpdateStatusInfo $new_data

				# evaluate post command in parent namespace...
				if {$widgetOptions(-xdeletepostcmd) != ""} {
					catch {uplevel $widgetOptions(-xdeletepostcmd)}
				}
				
				set widgetCompounds(undo_flag) 1
			}
		}
		
		my SelectionCallback
	}


	method EnableMoveover {} {
		my variable tblwidget
		
		# move-over effect:
		bind [$tblwidget bodypath] <Motion> {+
			set t [winfo parent %W]
			set x [expr {%x + [winfo x %W]}]
			set y [expr {%y + [winfo y %W]}]
			
			# set cell [$t nearestcell $x $y]
			# puts "Clicked on cell: $cell "
			# set rownum  [lindex [split $cell ","] 0]
			
			focus $t
			$t configure -activestyle frame
			$t activate "@$x,$y"
		}
		
		bind [$tblwidget bodypath] <Leave> {+
			set t [winfo parent %W]
			$t configure -activestyle none
		}
	}

}
